import datetime

from django.test import TestCase

from enumeration.const import ResetPeriod
from enumeration.models import Sequence, Counter, Gap
from enumeration.manager import consume_gap


class ConsumeGapTestCase(TestCase):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        cls.s = Sequence.objects.create(format='#', reset_period=ResetPeriod.DAILY)

    def test_no_counter(self):
        d = datetime.date(2012, 1, 14)
        self.assertIsNone(consume_gap(self.s, d))

    def test_has_counter_no_gaps(self):
        d = datetime.date(2012, 1, 14)
        Counter.objects.create(sequence=self.s, period=d, position=5)
        self.assertIsNone(consume_gap(self.s, d))

    def test_gaps_consumed(self):
        d = datetime.date(2012, 1, 14)
        c = Counter.objects.create(sequence=self.s, period=d, position=5)
        Gap.objects.create(counter=c, position=2)
        Gap.objects.create(counter=c, position=3)

        self.assertEqual(consume_gap(self.s, d), 2)
        self.assertFalse(Gap.objects.filter(counter=c, position=2).exists())

        self.assertEqual(consume_gap(self.s, d), 3)
        self.assertFalse(Gap.objects.filter(counter=c, position=3).exists())

    def test_right_counter_consumed(self):
        d = datetime.date(2012, 1, 14)
        c = Counter.objects.create(sequence=self.s, period=d, position=5)
        Gap.objects.create(counter=c, position=2)
        Gap.objects.create(counter=c, position=3)

        d2 = datetime.date(2012, 1, 15)
        c2 = Counter.objects.create(sequence=self.s, period=d2, position=5)
        Gap.objects.create(counter=c2, position=2)
        Gap.objects.create(counter=c2, position=3)

        self.assertEqual(consume_gap(self.s, d2), 2)
        self.assertEqual(c.gaps.count(), 2)
        self.assertEqual(c2.gaps.count(), 1)



import datetime
from typing import Optional, Tuple

from django.db import connection

from enumeration.const import ResetPeriod
from enumeration.models import Sequence
from .formatter import format_number


class EnumerationError(RuntimeError):
    pass


PERIOD_CRITERIA = "AND c.period = %s"

NEXT_QUERY = """
SELECT
  COALESCE("gap", "c"."position" + 1, 1) AS "position"
FROM enumeration_sequence s
LEFT JOIN enumeration_counter c
  ON c.sequence_id = s.id {}
LEFT JOIN LATERAL (
    SELECT
      "eg"."position" AS gap
    FROM enumeration_gap eg
    WHERE "eg"."counter_id" = c.id
    ORDER BY "eg"."position" ASC
    LIMIT 1) g ON true
WHERE s.id = %s;
"""


def truncate_date(period: ResetPeriod, date: datetime.date):
    if period == ResetPeriod.MONTHLY:
        return date.replace(day=1)

    if period == ResetPeriod.YEARLY:
        return date.replace(month=1, day=1)

    return date


def get_next_position(sequence: Sequence, date: Optional[datetime.date] = None) -> int:
    never = sequence.reset_period == ResetPeriod.NEVER

    args = [sequence.pk]

    if never:
        query = NEXT_QUERY.format("")

    elif date:
        query = NEXT_QUERY.format(PERIOD_CRITERIA)
        args = [truncate_date(sequence.reset_period, date)] + args
    else:
        raise EnumerationError("Missing date")

    with connection.cursor() as cursor:
        cursor.execute(query, args)
        row = cursor.fetchone()

    return row[0]


CONSUME_GAP_QUERY = """
DELETE FROM
    enumeration_gap
WHERE id = (
    SELECT
        g.id
    FROM enumeration_gap g
    INNER JOIN enumeration_counter c
        ON c.id = g.counter_id {}
    WHERE c.sequence_id = %s
    ORDER BY g."position" ASC LIMIT 1
) RETURNING "position";
"""


def consume_gap(sequence: Sequence, date: Optional[datetime.date] = None) -> Optional[int]:
    never = sequence.reset_period == ResetPeriod.NEVER

    args = [sequence.pk]

    if never:
        query = CONSUME_GAP_QUERY.format("")

    elif date:
        query = CONSUME_GAP_QUERY.format(PERIOD_CRITERIA)
        args = [truncate_date(sequence.reset_period, date)] + args

    else:
        raise EnumerationError("Missing date")

    with connection.cursor() as cursor:
        cursor.execute(query, args)
        row = cursor.fetchone()

    if row:
        return row[0]

    return None


def increment(sequence: Sequence, date: Optional[datetime.date] = None) -> int:
    never = sequence.reset_period == ResetPeriod.NEVER

    if never:
        query = """
         INSERT INTO enumeration_counter (sequence_id, "position") VALUES (%s, 1)
         ON CONFLICT (sequence_id) WHERE period IS NULL
         DO UPDATE SET "position" = enumeration_counter."position" + 1 RETURNING "position";
        """
        args = [sequence.pk]

    elif date:
        query = """
         INSERT INTO enumeration_counter (sequence_id, "position", period) VALUES (%s, 1, %s)
         ON CONFLICT (sequence_id, period)
         DO UPDATE SET "position" = enumeration_counter."position" + 1 RETURNING "position";
        """
        args = [sequence.pk, truncate_date(sequence.reset_period, date)]

    else:
        raise EnumerationError("Missing date")

    with connection.cursor() as cursor:
        cursor.execute(query, args)
        row = cursor.fetchone()

    return row[0]


# @TODO testcase
def get_number(sequence: Sequence, fill_gap=True, **context) -> Tuple[str, int]:
    date = context.get('date', None)
    if fill_gap:
        gap_pos = consume_gap(sequence, date)
        if gap_pos is not None:
            return format_number(sequence.format, position=gap_pos, **context), gap_pos

    pos = increment(sequence, date)
    return format_number(sequence.format, position=pos, **context), pos



from django.core.exceptions import ValidationError
from django.db import models
from enumfields import EnumField
from partial_index import PartialIndex, PQ

from enumeration.const import ResetPeriod
from enumeration.validators import validate_format


class Sequence(models.Model):
    format = models.CharField(max_length=255)
    reset_period = EnumField(ResetPeriod, max_length=30, default=ResetPeriod.NEVER)

    def clean(self):
        try:
            validate_format(self.format, self.reset_period)

        except ValidationError as e:
            raise ValidationError({
                "format": e.messages
            })


class Counter(models.Model):
    sequence = models.ForeignKey(Sequence, related_name='counters', on_delete=models.CASCADE)
    position = models.PositiveIntegerField(default=0)
    period = models.DateField(null=True)

    class Meta:
        unique_together = (
            ('sequence', 'period',)
        )
        indexes = [
            # only one counter for period-naive sequences
            PartialIndex(fields=['sequence'], unique=True, where=PQ(period__isnull=True)),
        ]


class Gap(models.Model):
    counter = models.ForeignKey(Counter, related_name='gaps', on_delete=models.CASCADE)
    position = models.PositiveIntegerField()

    unique_together = (
        ('counter', 'position',)
    )


# class EnumeratedDocumentMixin(models.Model):
#     sequence = models.ForeignKey(Sequence)
#     position = models.PositiveIntegerField(null=True)
#     number = models.CharField(max_length=255, null=True, blank=True)
#
#     class Meta:
#         abstract = True



from django.core.exceptions import ValidationError
from django.test import SimpleTestCase

from enumeration.const import ResetPeriod
from enumeration.validators import validate_format


class FormatValidatorTestCase(SimpleTestCase):
    def test_number_required(self):
        with self.assertRaises(ValidationError) as ctx:
            validate_format('', ResetPeriod.NEVER)

        self.assertTrue("Number placeholder \"#{1,10}\" is required." in ctx.exception.messages)

        validate_format('##', ResetPeriod.NEVER)
        validate_format('###', ResetPeriod.NEVER)

    def test_year(self):
        with self.assertRaises(ValidationError) as ctx:
            validate_format('#', ResetPeriod.YEARLY)

        self.assertTrue("Year placeholder \"Y{2,4}\" is required for yearly period." in ctx.exception.messages)

        with self.assertRaises(ValidationError) as ctx:
            validate_format('Y#', ResetPeriod.YEARLY)

        validate_format('YY###', ResetPeriod.YEARLY)

    def test_month(self):
        with self.assertRaises(ValidationError) as ctx:
            validate_format('#', ResetPeriod.MONTHLY)

        self.assertTrue("Year placeholder \"Y{2,4}\" is required for monthly period." in ctx.exception.messages)

        with self.assertRaises(ValidationError) as ctx:
            validate_format('YY#', ResetPeriod.MONTHLY)

        self.assertTrue("Month placeholder \"M{1,2}\" is required for monthly period." in ctx.exception.messages)

        validate_format('YYM#', ResetPeriod.YEARLY)
        validate_format('YYM###', ResetPeriod.YEARLY)

    def test_day(self):
        with self.assertRaises(ValidationError) as ctx:
            validate_format('#', ResetPeriod.DAILY)
        self.assertTrue("Year placeholder \"Y{2,4}\" is required for daily period." in ctx.exception.messages)

        with self.assertRaises(ValidationError) as ctx:
            validate_format('YY#', ResetPeriod.DAILY)
        self.assertTrue("Month placeholder \"M{1,2}\" is required for daily period." in ctx.exception.messages)

        with self.assertRaises(ValidationError) as ctx:
            validate_format('YYM#', ResetPeriod.DAILY)
        self.assertTrue("Day placeholder \"D{1,2}\" is required for daily period." in ctx.exception.messages)

        validate_format('YYMD#', ResetPeriod.YEARLY)
        validate_format('YYYYMD###', ResetPeriod.YEARLY)



import datetime

from django.test import TestCase

from enumeration.const import ResetPeriod
from enumeration.models import Sequence, Counter
from enumeration.manager import increment


class IncrementTestCase(TestCase):
    def test_never(self):
        s = Sequence.objects.create(format='#', reset_period=ResetPeriod.NEVER)
        self.assertEqual(increment(s), 1)
        # created
        c = Counter.objects.get(sequence=s, position=1, period=None)

        # updated
        self.assertEqual(increment(s), 2)
        self.assertTrue(Counter.objects.get(sequence=s, position=2, period=None, pk=c.pk))

    def test_daily_counter_created(self):
        s = Sequence.objects.create(format='#', reset_period=ResetPeriod.DAILY)

        d = datetime.date(2012, 1, 14)
        self.assertEqual(increment(s, d), 1)
        # created
        c = Counter.objects.get(sequence=s, position=1, period=d)

        # updated
        self.assertEqual(increment(s, d), 2)
        self.assertTrue(Counter.objects.get(sequence=s, position=2, period=d, pk=c.pk))

        d = datetime.date(2012, 1, 15)
        self.assertEqual(increment(s, d), 1)
        # new created
        Counter.objects.get(sequence=s, position=1, period=d)

    def test_monthly_counter_created(self):
        s = Sequence.objects.create(format='#', reset_period=ResetPeriod.MONTHLY)

        m = datetime.date(2012, 1, 1)
        self.assertEqual(increment(s, datetime.date(2012, 1, 14)), 1)
        # created
        c = Counter.objects.get(sequence=s, position=1, period=m)

        # updated
        self.assertEqual(increment(s, datetime.date(2012, 1, 14)), 2)
        self.assertTrue(Counter.objects.get(sequence=s, position=2, period=m, pk=c.pk))

        # next day, same month incremented
        next_d = datetime.date(2012, 1, 15)
        self.assertEqual(increment(s, next_d), 3)
        self.assertTrue(Counter.objects.get(sequence=s, position=3, period=m, pk=c.pk))

        self.assertEqual(increment(s, datetime.date(2012, 2, 1)), 1)
        # new created
        Counter.objects.get(sequence=s, position=1, period=datetime.date(2012, 2, 1))

    def test_yearly_counter_created(self):
        s = Sequence.objects.create(format='#', reset_period=ResetPeriod.YEARLY)

        y = datetime.date(2012, 1, 1)
        self.assertEqual(increment(s, datetime.date(2012, 1, 14)), 1)
        # created
        c = Counter.objects.get(sequence=s, position=1, period=y)

        # updated
        self.assertEqual(increment(s, datetime.date(2012, 1, 14)), 2)
        self.assertTrue(Counter.objects.get(sequence=s, position=2, period=y, pk=c.pk))

        # next month, same year incremented
        next_m = datetime.date(2012, 2, 14)
        self.assertEqual(increment(s, next_m), 3)
        self.assertTrue(Counter.objects.get(sequence=s, position=3, period=y, pk=c.pk))

        self.assertEqual(increment(s, datetime.date(2013, 2, 1)), 1)
        # new created
        Counter.objects.get(sequence=s, position=1, period=datetime.date(2013, 1, 1))



DATABASE_ENGINE = 'sqlite3'

DATABASES = {
    'default': {
        'ENGINE': 'django.db.backends.postgresql',
        'NAME': 'enumeration_test',
        'USER': 'postgres'
    }
}


INSTALLED_APPS = [
    # 'django.contrib.contenttypes',
    # 'django.contrib.sites',
    # 'django.contrib.auth',
    # 'django.contrib.admin',

    'enumeration',
]

SECRET_KEY = 'enumeration'



# Generated by Django 2.0.7 on 2018-07-13 12:55

from django.db import migrations, models
import django.db.models.deletion
import enumeration.const
import enumfields.fields
import partial_index
import partial_index.query


class Migration(migrations.Migration):

    initial = True

    dependencies = [
    ]

    operations = [
        migrations.CreateModel(
            name='Counter',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('position', models.PositiveIntegerField(default=0)),
                ('period', models.DateField(null=True)),
            ],
        ),
        migrations.CreateModel(
            name='Gap',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('position', models.PositiveIntegerField()),
                ('counter', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='gaps', to='enumeration.Counter')),
            ],
        ),
        migrations.CreateModel(
            name='Sequence',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('format', models.CharField(max_length=255)),
                ('reset_period', enumfields.fields.EnumField(default='never', enum=enumeration.const.ResetPeriod, max_length=30)),
            ],
        ),
        migrations.AddField(
            model_name='counter',
            name='sequence',
            field=models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='counters', to='enumeration.Sequence'),
        ),
        migrations.AddIndex(
            model_name='counter',
            index=partial_index.PartialIndex(fields=['sequence'], name='enumeration_sequenc_5647e6_partial', unique=True, where=partial_index.query.PQ(period__isnull=True), where_postgresql='', where_sqlite=''),
        ),
        migrations.AlterUniqueTogether(
            name='counter',
            unique_together={('sequence', 'period')},
        ),
    ]



import datetime

from django.test import TestCase

from enumeration.const import ResetPeriod
from enumeration.models import Sequence, Counter, Gap
from enumeration.manager import get_next_position, EnumerationError


class GetNextTestCase(TestCase):
    def test_no_reset(self):
        s = Sequence.objects.create(format='#', reset_period=ResetPeriod.NEVER)
        self.assertEqual(get_next_position(s), 1, "New sequence")

        c = Counter.objects.create(sequence=s, position=5)
        self.assertEqual(get_next_position(s), 6, "Counter position")

        Gap.objects.create(counter=c, position=3)
        self.assertEqual(get_next_position(s), 3, "Gap position")

    def test_daily_reset(self):
        s = Sequence.objects.create(format='#', reset_period=ResetPeriod.DAILY)

        # Date required
        with self.assertRaises(EnumerationError):
            get_next_position(s)

        d = datetime.date(2012, 1, 14)
        self.assertEqual(get_next_position(s, d), 1, "New sequence")

        c = Counter.objects.create(sequence=s, period=d, position=5)
        self.assertEqual(get_next_position(s, d), 6, "Counter position")

        Gap.objects.create(counter=c, position=3)
        self.assertEqual(get_next_position(s, d), 3, "Gap position")

        next_day = datetime.date(2012, 1, 15)
        self.assertEqual(get_next_position(s, next_day), 1, "New sequence")

        c = Counter.objects.create(sequence=s, period=next_day, position=5)
        self.assertEqual(get_next_position(s, next_day), 6, "Counter position")

        Gap.objects.create(counter=c, position=2)
        self.assertEqual(get_next_position(s, next_day), 2, "Gap position")

        # Yesterday doesn't change
        self.assertEqual(get_next_position(s, d), 3, "Gap position")

    def test_monthly_reset(self):
        s = Sequence.objects.create(format='#', reset_period=ResetPeriod.MONTHLY)

        # Date required
        with self.assertRaises(EnumerationError):
            get_next_position(s)

        d = datetime.date(2012, 1, 14)
        self.assertEqual(get_next_position(s, d), 1, "New sequence")

        c = Counter.objects.create(sequence=s, period=datetime.date(2012, 1, 1), position=5)
        self.assertEqual(get_next_position(s, d), 6, "Counter position")

        Gap.objects.create(counter=c, position=3)
        self.assertEqual(get_next_position(s, d), 3, "Gap position")

        next_day = datetime.date(2012, 1, 15)
        self.assertEqual(get_next_position(s, next_day), 3, "Same sequence")

        next_month = datetime.date(2012, 2, 1)
        self.assertEqual(get_next_position(s, next_month), 1, "New sequence")

        c = Counter.objects.create(sequence=s, period=datetime.date(2012, 2, 1), position=5)
        self.assertEqual(get_next_position(s, next_month), 6, "Counter position")

        Gap.objects.create(counter=c, position=3)
        self.assertEqual(get_next_position(s, next_month), 3, "Gap position")

    def test_yearly_reset(self):
        s = Sequence.objects.create(format='#', reset_period=ResetPeriod.YEARLY)

        # Date required
        with self.assertRaises(EnumerationError):
            get_next_position(s)

        d = datetime.date(2012, 1, 14)
        self.assertEqual(get_next_position(s, d), 1, "New sequence")

        c = Counter.objects.create(sequence=s, period=datetime.date(2012, 1, 1), position=5)
        self.assertEqual(get_next_position(s, d), 6, "Counter position")

        Gap.objects.create(counter=c, position=3)
        self.assertEqual(get_next_position(s, d), 3, "Gap position")

        next_day = datetime.date(2012, 1, 15)
        self.assertEqual(get_next_position(s, next_day), 3, "Same sequence")

        next_month = datetime.date(2012, 2, 1)
        self.assertEqual(get_next_position(s, next_month), 3, "Same sequence")

        next_year = datetime.date(2013, 2, 2)
        c = Counter.objects.create(sequence=s, period=datetime.date(2013, 1, 1), position=5)
        self.assertEqual(get_next_position(s, next_year), 6, "Counter position")

        Gap.objects.create(counter=c, position=3)
        self.assertEqual(get_next_position(s, next_year), 3, "Gap position")



from setuptools import setup, find_packages

setup(
    name='django-enumeration',
    version='0.1.0',
    url='https://github.com/shulcsm/django-enumeration',
    author='Mārtiņš Šulcs',
    author_email='shulcsm@gmail.com',
    packages=find_packages(exclude=['tests*']),
    zip_safe=False,
    include_package_data=True,
    license='MIT',
    classifiers=[
        'Environment :: Web Environment',
        'Framework :: Django',
        'Development Status :: 4 - Beta',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3.4',
        'Topic :: Software Development :: Libraries :: Application Frameworks',
        'Topic :: Software Development :: Libraries :: Python Modules',
    ],
    python_requires='>=3.6',
    install_requires=[
        'Django>=1.11',
        'django-partial-index==0.5.1',
        'django-enumfields==0.10.1'
    ]
)



from datetime import date
from django.test import SimpleTestCase
from enumeration.formatter import format_number


class FormatTestCase(SimpleTestCase):
    def test(self):
        self.assertEqual(format_number("#", position=1), '1')
        self.assertEqual(format_number("#", position=11), '11')
        self.assertEqual(format_number("##", position=1), '01')

        d = date(2017, 3, 11)

        self.assertEqual(format_number("#YY", position=1, date=d), '117')
        self.assertEqual(format_number("YY#", position=11, date=d), '1711')
        self.assertEqual(format_number("##YY", position=1, date=d), '0117')
        self.assertEqual(format_number("##YYYY", position=1, date=d), '012017')
        self.assertEqual(format_number("#YYMD", position=1, date=d), '117311')
        self.assertEqual(
            format_number("# {one} YY {two} M D", position=1, date=d, one='foo', two='bar'),
            '1 foo 17 bar 3 11'
        )



import re
from django.core.exceptions import ValidationError

from enumeration.const import ResetPeriod
from enumeration.formatter import POSITION, YEAR, MONTH, DAY


def validate_format(format_string, period: ResetPeriod):
    if not re.search(POSITION, format_string):
        raise ValidationError("Number placeholder \"#{1,10}\" is required.")

    if period in [ResetPeriod.YEARLY, ResetPeriod.MONTHLY, ResetPeriod.DAILY]:
        if not re.search(YEAR, format_string):
            raise ValidationError(f"Year placeholder \"Y{{2,4}}\" is required for {period.value} period.")

    if period in [ResetPeriod.MONTHLY, ResetPeriod.DAILY]:
        if not re.search(MONTH, format_string):
            raise ValidationError(f"Month placeholder \"M{{1,2}}\" is required for {period.value} period.")

    if period == ResetPeriod.DAILY:
        if not re.search(DAY, format_string):
            raise ValidationError(f"Day placeholder \"D{{1,2}}\" is required for {period.value} period.")



from typing import Match, Tuple, Pattern, Callable
import re
import datetime

POSITION = re.compile(r'#{1,10}')
YEAR = re.compile(r'Y{2,4}')
MONTH = re.compile(r'M{1,2}')
DAY = re.compile(r'D{1,2}')
CONTEXT = re.compile(r'{(\w+)}')


def _position(match: Match, position: int, **_context) -> str:
    return str(position).rjust(len(match.group()), '0')


def _year(match: Match, date: datetime.date, **_context) -> str:
    return str(date.year)[-len(match.group()):]


def _month(match: Match, date: datetime.date, **_context) -> str:
    return str(date.month).rjust(len(match.group()), '0')


def _day(match: Match, date: datetime.date, **_context) -> str:
    return str(date.day).rjust(len(match.group()), '0')


def _context(match: Match, **context) -> str:
    return context[match.group(1)]


FORMATTERS: Tuple[Tuple[Pattern, Callable], ...] = (
    (POSITION, _position),
    (YEAR, _year),
    (MONTH, _month),
    (DAY, _day),
    (CONTEXT, _context),
)


def format_number(format_string: str, **context) -> str:
    for pattern, resolve in FORMATTERS:
        for match in pattern.finditer(format_string):
            if match:
                format_string = format_string.replace(
                    match.group(), resolve(match, **context), 1)
    return format_string



import datetime

from django.db import IntegrityError
from django.test import TestCase

from enumeration.const import ResetPeriod
from enumeration.models import Sequence, Counter


class MiscTestCase(TestCase):
    def test_counter_constraint_same_date(self):
        s = Sequence.objects.create(format='#', reset_period=ResetPeriod.DAILY)
        Counter.objects.create(sequence=s, period=datetime.date(2012, 1, 14))

        with self.assertRaises(IntegrityError):
            Counter.objects.create(sequence=s, period=datetime.date(2012, 1, 14))

    def test_counter_constaint_nod_date(self):
        s = Sequence.objects.create(format='#', reset_period=ResetPeriod.NEVER)
        Counter.objects.create(sequence=s)

        with self.assertRaises(IntegrityError):
            Counter.objects.create(sequence=s)
